package org.mentaqueue.test.owt;

import java.util.Random;
import java.util.concurrent.locks.LockSupport;

import org.mentaaffinity.Affinity;
import org.mentaqueue.AtomicQueue;
import org.mentaqueue.BatchingQueue;
import org.mentaqueue.ConcurrentLinkedQueue;
import org.mentaqueue.util.Builder;
import org.mentaqueue.util.DetailedBenchmarker;
import org.mentaqueue.wait.SpinWaitStrategy;
import org.mentaqueue.wait.WaitStrategy;

/**
 * Results with this machine => http://mentablog.soliveirajr.com/lab/
 * 
 * More info here => http://mentablog.soliveirajr.com/2012/11/inter-thread-communication-with-2-digit-nanosecond-latency/
 * 
 * java -server -Xms1g -Xmx4g -XX:NewSize=512m -XX:MaxNewSize=1024m -cp target/classes/:../MentaAffinity/target/mentaaffinity.jar:../MentaLog/target/mentalog.jar:../MentaAffinity/lib/jna-3.5.1.jar org.mentaqueue.test.owt.ConcurrentLinkedQueueTest 100000000 100000000 0
 * 
 * Without inertia latency:
 * 
 * Iterations: 100000000 | Avg Time: 205.42 nanos | Min Time: 102 nanos | Max Time: 2385959 nanos | 75%: avg=195 max=222 | 90%: avg=201 max=235 | 99%: avg=204 max=252 | 99.9%: avg=205 max=266 | 99.99%: avg=205 max=543 | 99.999%: avg=205 max=2853
 * 
 * With inertia latency:
 * 
 * Iterations: 100000 | Avg Time: 242.36 nanos | Min Time: 159 nanos | Max Time: 32277 nanos | 75%: avg=232 max=257 | 90%: avg=237 max=269 | 99%: avg=240 max=293 | 99.9%: avg=241 max=465 | 99.99%: avg=241 max=1641 | 99.999%: avg=242 max=7103
 * 
 * @author Sergio Oliveira Jr.
 */
public class ConcurrentLinkedQueueTest {
	
	private static final int QUEUE_SIZE = 1024;
	private static final Random RANDOM = new Random();
	
	public static void main(String[] args) {
		
		final long messagesToWarmup = Long.parseLong(args[0]);
		final long messagesToTest = Long.parseLong(args[1]);
		final int delayBetweenMessages = Integer.parseInt(args[2]);
		
		final DetailedBenchmarker bench = new DetailedBenchmarker();
		
		final BatchingQueue<MutableLong> aToB = new ConcurrentLinkedQueue<MutableLong>(MutableLong.BUILDER);
		final AtomicQueue<MutableLong> bToA = new AtomicQueue<MutableLong>(QUEUE_SIZE, MutableLong.BUILDER);
		
		final WaitStrategy producerWaitStrategy = new SpinWaitStrategy();
		final WaitStrategy consumerWaitStrategy = new SpinWaitStrategy();
		
		Thread producer = new Thread(new Runnable() {
			
			private final void send(boolean warmup) {
				long ts = System.nanoTime(); // warmup this as well...
				MutableLong ml = aToB.nextToDispatch();
				ml.set(warmup ? 0 : ts);
				aToB.flush(); // no lazySet, send immediately
			}

			@Override
			public void run() {
				
				Affinity.bind();
				
				// first warmup...
				
				send(true); // send the first one to start!
				
				long count = 0;
				
				while(count < messagesToWarmup) {
					
					// receive echo to send the next message...
					long avail = bToA.availableToPoll();
					if (avail > 0) {
						bToA.poll();
						count++;
						bToA.donePolling(true); // can be lazy here because queue is not full...
						producerWaitStrategy.reset();
						if (count < messagesToWarmup) {
							send(true);
						}
					} else {
						producerWaitStrategy.waitForOtherThread();
					}
				}
				
				// now benchmark...
				
				send(false); // send the first one to start!
				
				count = 0;
				
				while(count < messagesToTest) {
					
					// receive echo to send the next message...
					long avail = bToA.availableToPoll();
					if (avail > 0) {
						bToA.poll();
						count++;
						bToA.donePolling(true); // can be lazy here because queue is not full...
						producerWaitStrategy.reset();
						if (count < messagesToTest) {
							if (delayBetweenMessages == 0) {
								send(false);
							} else if (delayBetweenMessages < 0) {
								LockSupport.parkNanos(RANDOM.nextInt(-1 * delayBetweenMessages));
								send(false);
							} else {
								LockSupport.parkNanos(delayBetweenMessages);
								send(false);
							}
						}
					} else {
						producerWaitStrategy.waitForOtherThread();
					}
				}
				
				Affinity.unbind();
				
				// DONE!
				System.out.println(bench.results());
				
			}
		}, "Thread-Producer");
		
		Thread consumer = new Thread(new Runnable() {

			@Override
			public void run() {
				
				Affinity.bind();
				
				while (true) {
					
					// time and echo back...
					
					long avail = aToB.availableToPoll();
					if (avail > 0) {
						MutableLong ml = aToB.poll();
						long ts = ml.get();
						aToB.donePolling(); // can be lazy because queue will never be full...
						if (ts > 0) {
							bench.measure(System.nanoTime() - ts);
						}
						
						consumerWaitStrategy.reset();
						
						// echo back
						MutableLong back = bToA.nextToDispatch();
						back.set(ts);
						bToA.flush(); // send immediately so the producer gets the echo...
					} else {
						consumerWaitStrategy.waitForOtherThread();
					}
				}
			}
		}, "Thread-Consumer");
		
		if (Affinity.isAvailable()) {
			Affinity.assignToProcessor(2, producer);
			Affinity.assignToProcessor(3, consumer);
		} else {
			System.err.println("Thread affinity not available!");
		}
		
		producer.setDaemon(false);
		consumer.setDaemon(true);

		consumer.start();
		try { Thread.sleep(1); } catch(Exception e) { }
		producer.start();

	}
		
	private static class MutableLong {
		
		private long value = 0L;

		public MutableLong(long value) {
			this.value = value;
		}

		public final long get() {
			return value;
		}

		public final void set(long value) {
			this.value = value;
		}
		
		@Override
		public String toString() {
			return String.valueOf(value);
		}
		
		public final static Builder<MutableLong> BUILDER = new Builder<MutableLong>() {
			@Override
	        public MutableLong newInstance() {
		        return new MutableLong(-1);
	        }
		};
	}
}